	; IMPORT 1.3	Imports a CP/M file from TPDD emulator
	;				
	; For M100 CP/M

	; By Philip Avery	1.3   May 2020
	;			1.0   Oct 2018
	; 			extension to 8080 by SA / Aug 2020

	.org	7FEAh		; Actually at 100h
	.include "common addresses.asm"
		
	
	lxi	b,Import_end-Import2	; Move program to upper RAM (8000h) so System ROM is always available
	lxi	d,8000h
	lxi	h,116h
Import1	mov	a,m
	stax	d
	inx	d
	inx	h
	dcx	b
	mov	a,b
	ora	c
	jnz	109h		; Which is Import1
	jmp	8000h		; Main program org 8000h
	
start:				; actual code start at 8000h

Import2	lxi	sp,stack	; Use our stack
				; Format is IMPORT M100.DA CPMFILE1.DAT (6.2 8.3)

	lda	5dh		; Any filename given?
	cpi	' '
	jz	Import5		; No, Import help

	mvi	c,6		; Save M100 file name (6.2 Source filename)
	lxi	d,5dh
	lxi	h,INPUT_BUFFER	; Filename for TPDD action
	call	Copy_de2hl
	mvi	a,2eh		; Add '.'
	mov	m,a
	inx	h
	mvi	c,2
	lxi	d,65h
	call	Copy_de2hl
	mvi	c,17		; Pad filename with whitespace & append with 46 00
	lxi	d,append
	call	Copy_de2hl

				; Process Destination filename

	lda	6dh		; First char of filename
	cpi	' '		; Is it blank?
	jnz	Imprt28		; No, full filename given
					
	
Imprt21	mvi	c,8		; Using source filename
	lxi	d,5dh
	lxi	h,6dh
	call	Copy_de2hl

	lda	75h		; Was there a dest. extension given?
	cpi	' '
	jnz	Imprt28		; Yes, we now have full filename
	
	mvi	c,3		; No, use source extension
	lxi	d,65h
	lxi	h,75h
	call	Copy_de2hl

Imprt28	mvi	c,16		; Full filename for dest. Move CP/M file name to FCB
	lxi	d,6ch
	lxi	h,5ch
	call	Copy_de2hl
	
Imprt29	lxi	d,5ch		; File FCB. Establish CP/M file
	mvi	c,19
	call	Bdos		; Delete File
	lxi	d,5ch
	mvi	c,22
	call	Bdos		; Make File
	lxi	d,nodir
	inr	a
	jz	Error_exit
	lxi	d,length_addr+1	; Make Input Buffer the DMA for writes
	mvi	c,26
	call	Bdos
	
	lxi	d,msgt6		; Print "Importing"
	mvi	c,9		; Print message
	call	Bdos
	
	call	Opt_off
	call	setup_TPDD	; Get file from TPDD...
	call	TPDD_prep_open
	lxi	h,INPUT_BUFFER
	mov	a,m
	sta	TEMP3		; store first character of name
	mvi	a,03h
	mov	m,a		; place mode here

	LXI 	H,0101H		; load HL with 0101		
	CALL	SEND_TPDD_COMMAND	; call - send command OPEN file
	lxi	h,INPUT_BUFFER
	lda	TEMP3
	mov	m,a		; restore first character of name
	
Import3	call	Opt_off
	Call	get_128_bytes	; Get a record of TPDD file
	call	Opt_on

	lda	length_addr	; Is record a full one (128-bytes long)?
	ora	a
	jz	Close		; No more sectors to read, EOF
	mov	c,a
	mvi	a,80h
	sub	c
	jz	Write_record	; Yes, write full record
	lxi	h,length_addr+1	; No, set rest of record to 00
	mvi	b,0
	dad	b	
	mov	c,a		; Number of bytes to set to zero
	xra	a
Import4	mov	m,a		; Start of 00
	inx	h
	dcr	c
	jnz	Import4

Write_record			; Write record to CP/M
	lxi	d,5ch		; File FCB
	mvi	c,21
	call	Bdos		; Write record
	lxi	d,dskfull	; Possible error message ready
	ora	a
	jnz	Error_exit
	lda	length_addr	; Was record a full one (128-bytes long)?
	sui	80h
	jm	Close		; No, only partial, so close
	
	lda	dotcnt		; Need to display a progress dot?
	dcr	a
	sta	dotcnt
	jnz	Wr1
	mvi	a,8
	sta	dotcnt
	mvi	e,'.'		; Print a dot to show progress
	mvi	c,2
	call	Bdos
Wr1	jmp	Import3

	
Import5	lxi	d,help_message
	call	Opt_on
	mvi	c,9		; Print help message
	call	Bdos
	jmp	Exit		; Quit

Close	pop	b		; Fix stack from EOF 3FH detection
	call	Opt_off
	call	TPDD_finish
	call	Opt_on
	lxi	d,5ch		; Yes, finish writing & close file
	mvi	c,16
	call	Bdos
Exit	jmp	0		; Boot
	
	

Copy_de2hl:			; For C bytes
	ldax	d
	inx	d
	mov	m,a
	inx	h
	dcr	c
	jnz	Copy_de2hl
	ret

Error_exit:
	pop	b		; Fix stack from EOF 3FH detection
	call	Opt_on
	mvi	c,9		; Print relevant error message
	call	Bdos
	jmp	Exit		; Quit
					
Opt_off	lda     0FF45h		; Turn Off Option ROM (lower 32k RAM), ie enable System ROM
       	ani     0FEh            ; Manage FF45 properly
       	sta     0FF45h
       	out     0E8h
	ret

Opt_on	lda	0FF45h	; Turn On Option ROM (lower 32k RAM)
	ori	01h
	sta	0FF45h
	out	0E8h
	ret

get_128_bytes:
	LXI 	H, 0003H		; load HL with 0003 read file
	CALL	SEND_TPDD_COMMAND	; call - send command	
					; at this point, file data has been moved to F6B7
					; F6B7 is a buffer
	xchg				; de = start of payload data
	ret

;-----------------------------------------------------------------------

setup_TPDD:
	call	set_DTR
	call	set_parameters
	ret

;-----------------------------------------------------------------------

TPDD_prep_open:
					;filename is prepped for use to "open file"								
	LXI 	H,0007H			;load HL with 0007 drive status					
	CALL	SEND_TPDD_COMMAND	;call - send command

	LXI 	H,1A00H			;load HL with 1A00 find file					
	CALL	SEND_TPDD_COMMAND	;call - send command
	ret
;-----------------------------------------------------------------------
TPDD_finish:
					
	LXI 	H,0002H			;load HL with 0002 close file				
	CALL	SEND_TPDD_COMMAND	;call - send command								
	ret

												
;-------------------------------------------------------------------------------												
;This routine sends the command, and loads the response from disk							
;-------------------------------------------------------------------------------
												
SEND_TPDD_COMMAND:	

	di
					;input HL = first 2 bytes after ZZ  - Z Z L H
					;send out command bytes, stored in input buffer at F685
					;wait for response
					;place all response bytes into buffer
					;returns with
					;load b with first response byte
					;load a with 3rd response byte
					;zero bit zet if a=0, reset if a <>0												
	mvi	c,00h			;zero out C
					;these bytes hold the first 2 bytes following ZZ
	LXI 	D,INPUT_BUFFER-2d	;load DE with F683
; 8080 change
;was-----------
;	SHLX				;store HL at F683=L F684=H (LH 0700 ,001A, 0101, 0300, 0200)
					;command data stored at beginning of command text
;------------
;replaced with

	mov	a,l
	stax	d
	inx	d
	mov	a,h
	stax	d
	dcx	d	
;--------------			
					
	INR 	H			;increment H								
	INR 	H			;increment H; add 2 bytes to loop counter
					;space for "F" and directory reference
					;h holds the number of command bytes (h = length not including ZZ)								
	MVI 	A, 5AH			;load A with 5A "Z"
	call	send_a_byte
	call	send_a_byte		;this routine preserves hl

snd_com_loop:													
	LDAX 	D			;load A from contents of DE	
	call	send_a_byte
												
	ADD 	C			;add C to A		a=contents of memory at DE - add A to current checksum							
	MOV 	C,A			;put in C		update checksum							
	INX 	D			;increment DE		move to next byte							
	DCR 	H			;decrement h - h holds the number of characters							
	JNZ	snd_com_loop		;loop back until h characters have been sent									

	xchg				; de points to first response byte location in buffer
	lxi	b,0020h
	dad	b			; add space for name in command location
	shld	TEMP1			; TEMP1 holds next byte location
	xchg

	inx	d			; de points to second incoming byte	

	CMA				; complement A - A holds the checksum
	call	send_a_byte													


; start receive function here

; checksum of the response is the ones compliment of the lsb of the sum
; of all bytes excluding the checksum

; sum them up, subtract the checksum (last byte), xor FF  (ones compliment)  compare
; store checksum at the "next byte" location

					; command is sentnow get response

	lhld	TEMP1
	xra	a
	mov	m,a			; store 00 at TEMP1 (starting checksum)

get_response_loop:

get_serial_byte:

	call	Get_char
	lhld	TEMP1			; get location for next receive character
					; should hold checksum

	mov	c,m			; get checksum
	mov	m,a			; store data
	add	c			; add to checksum

	inx	h			; hl points to next character location
	mov	m,a			; store new checksum
	shld	TEMP1			; store new next character


	ldax	d			; get byte from de location
					; hl holds next byte location
	inr	a
	inr	a			; add 2 to a
; 8080 change	
;was------------------------
;	mov	c,a
;	mvi	b,00h		
;	dsub				; subtract length bytes from location		
					; end when result = de
;is now---------------------
					
	cma				; because hl + complement(00aa) + 1 = hl - 00aa
	mov	c,a			; test with aa = 00
	mvi	b,0FFh			; hl + complement(0000) + 1 = hl + FFFF + 1 = hl = hl - 0000
	dad	b
	inx	h
					
;----------------------------					
					
					
	mov	a,h
	cmp	d
	
	jnz	get_response_loop

	mov	a,l
	cmp	e
	
	jnz	get_response_loop

;  1   2   3   4   5
; XX  02  YY  YY  CS
; a1  a2  a3  a4  a5  HL
;     DE

; now compare the checksums

	lhld	TEMP1
	mov 	a,m			; load a with checksum
	dcx	h
	sub	m			; subtract response checksum value (not included in total)
					; a holds actual checksum
	xri	0FFH			; toggle all bits
	cmp	m			; compare to response checksum at m
	jnz	response_checksum_error

	xchg
	dcx	h			;get back original start of response

	MOV 	B,M			;load B from M load B with first received character
	INX 	H			;increment HL				
	INX 	H			;increment HL skip ahead two bytes
	xra	a
	ADD 	M			;add M to A load A with M (A was zero)
					;return with zero bit set if M=0

					; check for an error response
					; hl points to start of payload of response

;----------------------------------------------------------------------
disk_error:			; disk error display
				; if reponse byte is 12 then an error is possible
	mov	c,a
	push	b		; temp store a
	mov	a,b
	sui	12h		; see if b=12h
	pop	b		
	mov	a,c		; restore a
	rnz			; return if not an possible error response

				; there was a disk error
				; if a = 0 then return - no error
	ora	a		; check if a = 0
	rz			; return from read response routine if no error

	cpi	3fh		; Is error End of file?
	jz	Close		; Yes, then close. NB stack bad

				; there was an error!
	ani	11110000b		; and with F0
	rrc				; error in range 10-80
	rrc
	rrc				; rotate right 3 times, we want 2x 
	
	lxi	h, error_table
	mov	c,a
	mvi	b,00h
	dad	b			; offset hl with error reference hl = hl + 2*a
					; hl points to error message
	xchg				; place pointer in de

;8080 change
;was----------------------------
;	lhlx				; hl = (DE)
;is now-------------------------	

	push	psw
	ldax	d
	mov	l,a
	inx	d
	ldax	d
	mov	h,a
	dcx	d
	pop	psw
;-------------------------------	
	
	
	
	
	xchg				; de points to error message		
	jmp	Error_exit		; there was a disk error, so bail out!

error_table:
	.dw	unspecified_error
	.dw	file_name_error
	.dw	unknown_error
	.dw	command_error
	.dw 	read_error
	.dw	write_protect
	.dw	disk_full
	.dw	insert_disk
	.dw	hardware_error


errors:
unspecified_error:
	.db	"Unspecified error$"		; code 0xh
file_name_error:
	.db	"File name error$"		; code 1xh
unknown_error:
	.db	"Unknown error$"		; code 2xh
command_error:
	.db	"Command error$"		; code 3xh
read_error:
	.db	"Read error$"			; code 4xh
write_protect:
	.db	"Write protected$"		; code 5xh
disk_full:
	.db	"Disk full$"			; code 6xh
insert_disk:
	.db	"Insert disk$"			; code 7xh
hardware_error:
	.db	"Hardware error$"		; code 8xh

;00 - normal (no error)
;10 - file does not exist 11 - file exists
;30 - no filename 31 - dir search error 35 - bank error
;36 - parameter error 37 - open format mismatch 3f - end of file
;40 - no start mark 41 - crc check error in ID 42 - sector length error
;44 - format verify error 46 - format interruption 47 - erase offset error
;49 - crc check error in data 4a - sector number error 4b - read data timeout
;4d - sector number error 50 - disk write protect 5e - un-initilized disk
;60 - directory full 61 - disk full 6e - file too long
;70 - no disk 71 - disk change error


;----------------------------------------------------------------------
Load_failed_error:			; error when load does not succeed
	lxi	d,msgt10
	jmp	Error_exit
;----------------------------------------------------------------------
TPDD_response_error:			; error when TPDD does not respond
	lxi	d,msgt11
	jmp	Error_exit
;----------------------------------------------------------------------				
send_error:				; print error when esc hit
	lxi	d,msgt9
	jmp	Error_exit
;----------------------------------------------------------------------				
response_checksum_error:		; print error when checksum bad
	lxi	d,msgt7
	jmp	Error_exit

;---------------------------------------------------------------------
Get_char:
	ei
	call	6D6Dh
	jz	Get_char		; Wait for char
	di
	jmp	6D7Eh
;---------------------------------------------------------------------
set_parameters:
	mvi	a,0FFh
	adi	01h			; set carry for RS232
	LXI 	H,SERPAR		; POINT HL TO SERIAL PARAMETERS	
;	rst	6
	call	017E6h			; Set RS232 parameters from string at M
	ret
SERPAR	.db	"98N1D"

;------------------------------------------------------------------
set_DTR:
	in 	0BAh
	ani	10111111b		; assert DTR,
	ori	10000000b		; de assert RTS
	out	0BAh			; bit 7 = RTS, bit 6 = DTR
	ret				; IGNORE DSR




;---------------------------------------------------------------------------------
;TPDD errors
msgt6:		.db	"Importing",0Dh,0Ah,"$"
msgt7:		.db	"checksum error...aborting$"
msgt9:		.db	"error sending command$"
msgt10:		.db	"load file from TPDD failed$"
msgt11:		.db	"no response from TPDD$"

INPUT_BUFFER	.equ	0F685h 
TEMP1		.dw	0000
TEMP3		.db	00
TEMP5		.dw	0000
send_a_byte	.equ	6E32h

append		.db 	"               ",46h,00
dotcnt		.db	8	; Dot count - progress 8 sectors = 1KB
help_message	.db	"IMPORT 1.3 imports file from TPDD-emulator into CP/M. 6.2...8.3 format Eg:",0ah,0ah,0dh
		.db	"import file12.da file1234.dat",0ah,0dh
		.db	"import file12.da .dat (saves file12.dat",0dh,0Ah
		.db	"import file12.da      (saves file12.da)$"

length_addr	.equ	0F6A6h	; TPDD data length byte
	
		.ds	100	; Allow for M100 ROM interrupts
stack
nodir		.db	"No Directory space$"
dskfull		.db	"Disk full$"

Import_end

	.end